(*
   Title  : Fold_Parameters.thy
   Author : M. Bortin
*)

theory Fold_Parameters
imports QuotientType
begin

section {* Auxiliary concepts *}

constdefs
 division :: "('a \<times> 'b) set \<Rightarrow> ('a \<times> 'b) set \<Rightarrow> 'a set" (infixr "\\" 55)
"(R \\ S) \<equiv> {a. \<forall>b. (a, b) \<in> R \<longrightarrow> (a, b) \<in> S}"

 ker :: "('a \<Rightarrow> 'b) \<Rightarrow> ('a \<times> 'a) set"
"ker f \<equiv> {(a, b). f a = f b}"

lemma congruence_factor :
"congruence f = (UNIV \<subseteq> eqR \\ ker f)"
apply(simp add: division_def ker_def)
apply(simp add: congruence_def)
apply(rule iffI)
apply clarsimp
apply clarsimp
apply(drule_tac c="s" in subsetD, simp)
apply simp
done



section {* The formal parameters *}


typedecl ('u, 'a) \<Sigma>

consts
 \<Sigma> :: "('a \<Rightarrow> 'b) \<Rightarrow> ('u, 'a) \<Sigma> \<Rightarrow> ('u, 'b) \<Sigma>"
 \<Sigma>' :: "('a \<times> 'b) set \<Rightarrow> (('u, 'a) \<Sigma> \<times> ('u, 'b) \<Sigma>) set" 


text{* The extension of @{term "eqR"} to \<Sigma> *}
constdefs
 eqR_ext :: "('a, 'a T) \<Sigma> \<times> ('a, 'a T) \<Sigma> \<Rightarrow> bool"   
"eqR_ext \<equiv> \<Sigma>' eqR"

syntax
 eqR_ext_ :: "('a, 'a T) \<Sigma> \<Rightarrow> ('a, 'a T) \<Sigma> \<Rightarrow> bool"    ("_/ \<simeq>\<^sub>\<Sigma> _"  [50, 51] 50)
translations
"a \<simeq>\<^sub>\<Sigma> b" == "(a, b) \<in> eqR_ext"


text{* Embedding and projection of monotypes *}
constdefs
 monoE :: "'a set \<Rightarrow> ('a \<times> 'a) set"
"monoE X \<equiv> {(a, b). a \<in> X \<and> a = b}"

 monoP :: "('a \<times> 'a) set \<Rightarrow> 'a set"
"monoP Y \<equiv> {a. (a, a) \<in> Y}"


text{* The action of \<Sigma> on monotypes *}
consts
 \<Sigma>'' :: "'a set \<Rightarrow> (('u, 'a) \<Sigma>) set" 
defs
 Sigma''_def : "\<Sigma>'' \<equiv> monoP \<circ> \<Sigma>' \<circ> monoE"


axioms
 Sigma_division : "\<Sigma>'' (eqR \\ ker f) \<subseteq> eqR_ext \\ ker (\<Sigma> f)" 



subsection {* A \<Sigma>-coalgebra as a formal parameter *}

consts
 TC :: "'a T \<Rightarrow> ('a, 'a T) \<Sigma>"

axioms
 TC_reduct : "TC -` (\<Sigma>'' P) \<subseteq> P \<Longrightarrow> UNIV \<subseteq> P"



subsection {* A hylomorphism parameter: Fold *}

consts
 Fold     :: "(('a, 'b) \<Sigma> \<Rightarrow> 'b) \<Rightarrow> 'a T \<Rightarrow> 'b"
 FoldCond :: "(('a, 'b) \<Sigma> \<Rightarrow> 'b) set"

axioms
 Fold_eq        : "A \<in> FoldCond \<Longrightarrow> Fold A = A \<circ> \<Sigma> (Fold A) \<circ> TC" 


subsection {* Transformation *}

consts 
 Transform :: "('a, 'a T) \<Sigma> \<Rightarrow> ('a, 'a T) \<Sigma> \<Rightarrow> ('a, 'a T) \<Sigma>"
 TransformCond :: "(('a, 'b) \<Sigma> \<Rightarrow> 'b) \<Rightarrow> bool"

axioms
 Transform_eqR   : "s \<simeq> t \<Longrightarrow> TC s \<simeq>\<^sub>\<Sigma> Transform (TC s) (TC t)" 
 Fold_Transform  : "A \<in> TransformCond \<Longrightarrow> s \<simeq> t \<Longrightarrow> A (\<Sigma> (Fold A) (Transform (TC s) (TC t))) = Fold A t"
 TransformCond   : "TransformCond \<subseteq> FoldCond"



end