
theory QuotientType
imports QuotientType_Parameters 
begin

typedef 'a TQ = "(UNIV // eqR) :: ('a T set) set"
(*<*)
apply(rule_tac x="eqR `` {a}" in exI)
apply(rule quotientI)
by simp
(*>*)

lemma eqR_equiv : "equiv UNIV eqR"
(*<*)
apply(simp add: equiv_def refl_on_def sym_def trans_def)
apply(intro conjI)
apply(clarify, rule eqR_refl)
apply(clarify, erule eqR_sym)
by(clarify, erule eqR_trans) 
(*>*)



text 
{* Class operations. *}

constdefs
 repr_of :: "'a TQ \<Rightarrow> 'a T"        ("^_" [99] 1000)
"repr_of q \<equiv> (\<some> a. a \<in> Rep_TQ q)"

 class_of :: "'a T \<Rightarrow> 'a TQ"       ("<_>" [99] 1000)
"class_of t \<equiv> Abs_TQ (eqR `` {t})"

lemma class_of_eq : "(s \<simeq> t) = (<s> = <t>)"
(*<*)
apply(simp add: class_of_def, rule iffI)
apply(rule Rep_TQ_inject[THEN iffD1])
apply(subst Abs_TQ_inverse)
apply(simp add: TQ_def)
apply(rule quotientI, simp)
apply(subst Abs_TQ_inverse)
apply(simp add: TQ_def)
apply(rule quotientI, simp)
apply(rule equiv_class_eq)
apply(rule eqR_equiv)
apply(assumption)
apply(drule Rep_TQ_inject[THEN iffD2])
apply(subst (asm) Abs_TQ_inverse)
apply(simp add: TQ_def)
apply(rule quotientI, simp)
apply(subst (asm) Abs_TQ_inverse)
apply(simp add: TQ_def)
apply(rule quotientI, simp)
apply(drule_tac f="\<lambda>x. t \<in> x" in arg_cong)
apply(drule iffD2)
apply(simp add: Image_def)
apply(rule eqR_refl)
apply(simp add: Image_def)
done
(*>*)


lemma class_of_repr : "<^q> = q"
(*<*)
apply(simp add: class_of_def)
apply(rule Rep_TQ_inject[THEN iffD1])
apply(subst Abs_TQ_inverse)
apply(simp add: TQ_def)
apply(rule quotientI, simp)
apply(insert Rep_TQ[of q])
apply(simp add: TQ_def)
apply(erule quotientE)
apply(rule_tac t="Rep_TQ q" in ssubst, assumption)
apply(rule equiv_class_eq)
apply(rule eqR_equiv)
apply(simp add: repr_of_def)
apply(rule_tac a=x in someI2)
apply(rule eqR_refl)
apply(erule eqR_sym)
done
(*>*)


lemma repr_of_class : "^<t> \<simeq> t" 
(*<*)
apply(simp add: repr_of_def)
apply(rule_tac a=t in someI2)
apply(simp add: class_of_def)
apply(subst Abs_TQ_inverse)
apply(simp add: TQ_def)
apply(rule quotientI, simp)
apply(simp add: Image_def, rule eqR_refl)
apply(simp add: class_of_def)
apply(subst (asm) Abs_TQ_inverse)
apply(simp add: TQ_def)
apply(rule quotientI, simp)
apply(simp add: Image_def)
by(erule eqR_sym)
(*>*)


lemma repr_inj : "^s = ^t \<Longrightarrow> s = t"
(*<*)
apply(subst class_of_repr[THEN sym])
apply(subst class_of_repr[THEN sym, of "t"])
apply(subst class_of_eq[THEN sym])
apply(erule subst)
by (rule eqR_refl)
(*>*)


text 
{* Congruences and extension. *}


constdefs
 congruence :: "('a T \<Rightarrow> 'b) \<Rightarrow> bool"
"congruence f \<equiv> (\<forall>s t. s \<simeq> t \<longrightarrow> f s = f t)"

 qtype_ext :: "('a T \<Rightarrow> 'b) \<Rightarrow> ('a TQ \<Rightarrow> 'b)"  ("_\<^sup>#" [99] 1000)
"f\<^sup># \<equiv> f \<circ> repr_of"




lemma congruence_comp : 
"congruence g \<Longrightarrow> congruence (f \<circ> g)"
(*<*)
by(simp add: congruence_def)
(*>*)


lemma ext1 : 
"congruence f \<Longrightarrow> f\<^sup># <s> = f s"
(*<*)
apply(simp add: qtype_ext_def congruence_def)
apply(drule spec)+
apply(erule mp)
apply(rule repr_of_class)
done
(*>*)

corollary ext2 : 
"congruence (class_of \<circ> f) \<Longrightarrow> (class_of \<circ> f)\<^sup># <s> = <f s>"
by(subst ext1, assumption, simp)



lemma congruence_req : "(\<forall>s. f\<^sup># <s> = f s) \<Longrightarrow> congruence f"
(*<*)
apply(simp add: congruence_def, clarify)
apply(frule_tac x=s in spec)
apply(drule_tac x=t in spec)
apply(erule subst)
apply(erule subst)
apply(subst (asm) class_of_eq)
apply(erule subst)
by(rule refl)
(*>*)



lemma ext_comp :
"congruence (class_of \<circ> f) \<Longrightarrow> congruence g \<Longrightarrow>
 g\<^sup># ((class_of \<circ> f)\<^sup># q) = (g \<circ> f)\<^sup># q"
(*<*)
apply(subst class_of_repr[of q, THEN sym])
apply(subst ext2, assumption)
apply(subst ext1, assumption)
by(simp add: qtype_ext_def)
(*>*)


lemma congruence_comp2 : 
"congruence (class_of \<circ> g) \<Longrightarrow> congruence (class_of \<circ> f) \<Longrightarrow>
 congruence (class_of \<circ> f \<circ> g)"
(*<*)
apply(simp (no_asm) add: congruence_def)
apply clarify
apply(subst ext2[where f=f, THEN sym], assumption)+
apply(rule_tac f="(class_of \<circ> f)\<^sup>#" in arg_cong)
apply(subst ext2[where f=g, THEN sym], assumption)+
apply(rule_tac f="(class_of \<circ> g)\<^sup>#" in arg_cong)
apply(erule class_of_eq[THEN iffD1])
done
(*>*)


lemma ext_property : 
"f s \<simeq> t \<Longrightarrow> congruence (class_of \<circ> f) \<Longrightarrow> (class_of \<circ> f)\<^sup># <s> = <t>"
(*<*)
apply(subst ext2, assumption)
by(erule class_of_eq[THEN iffD1])
(*>*)


lemma quotient_induct[rule_format] :
"\<forall>t. P <t> \<Longrightarrow> P q"
apply(subst class_of_repr[THEN sym]) 
by simp

end


